/* Scarab - provides SLIP over a half-duplex link such as infrared or radio.
 *
 * kernel.c
 * Allocates a pty pair, connects one side to the kernel's SLIP
 * interface and returns the other side to the caller.
 *
 * (c) Copyright 1996 Alan Cox, Robin O'Leary <robin@acm.org>.
 * All Rights Reserved.
 *
 * This program is distributed without any warranty; without even the
 * implied warranty of MERCHANTABILITY or fitness for a particular purpose.
 * See the file LICENSE for details.
 */

#include <netdb.h>	/* gethostbyname() */
#include <stdio.h>	/* fprintf() */
#include <arpa/inet.h>	/* inet_addr() */
#include <fcntl.h>	/* open() */
#include <unistd.h>	/* close() */
#include <sys/ioctl.h>	/* for ioctl() prototype */

/* Would like to include this, but it collides with gcc
 * #include <linux/if_slip.h>	/* SL_MODE_SLIP */
#define SL_MODE_SLIP 0
#define sockaddr _sockaddr
#define linger _linger
#define msghdr _msghdr
#include <linux/if.h>	/* IFF_UP etc, struct ifr */

#include "main.h"	/* for global variables */
#include "debug.h"	/* for DEBUG_KERNEL */
#include "term.h"	/* for set_ternios() */
#include "slip.h"	/* for slip_write_frame() */
#include "scarab.h"	/* for scarab_send_slip_frame() */


int kernel_fd=-1;	/* set by kernel_start_slip() */

typedef unsigned int uint32;

/****************************** IP Name handling ***************************/

static int get_ip_addr(const char *ptr, uint32 *addr)
{
/*
	long l;
	if((l=inet_addr(ptr))==-1)
	{
		fprintf(stderr,"%s: %s is not a legal IP address.\n", ptr);
		perror("inet_addr");
		return -1;
	}
	*addr=(unsigned long)l;
*/
	struct hostent *h=gethostbyname(ptr);
	if(!h)
	{
		fprintf(stderr,"%s: can't resolve hostname %s\n",
			program_name, ptr);
		herror("gethostbyname");
		return -1;
	}
	/* h_addr is actually char[], but we presume that for AF_INET
	 * addresses that'll be four bytes which we can cast.
	 */
	*addr=*( (uint32 *)h->h_addr );
	return 0;
}

/******************************* Pty Allocation ****************************/

static int ptypair(int *pty, int *tty)
{
	char ttyname[]="/dev/ttyXX";
	char ptyname[]="/dev/ptyXX";
	char tmplt1[]="pqrstuvwxyz";
	char tmplt2[]="0123456789abcdef";
	
	int  tc1,tc2;
	int  pfd,tfd;
	
	for(tc1=0;tc1<11;tc1++)
	{
		for(tc2=0;tc2<16;tc2++)
		{
			ptyname[8]=tmplt1[tc1];
			ptyname[9]=tmplt2[tc2];
			pfd=open(ptyname, O_RDWR|O_NDELAY);
			if(pfd==-1)
				continue;
			ttyname[8]=tmplt1[tc1];
			ttyname[9]=tmplt2[tc2];
			tfd=open(ttyname, O_RDWR|O_NDELAY);
			if(tfd==-1)
			{
				close(pfd);
				continue;
			}
			fcntl(pfd, F_SETFL, 0);
			fcntl(tfd, F_SETFL, 0);
			
			set_termios(pfd, 38400);
			set_termios(tfd, 38400);
			
			*pty=pfd;
			*tty=tfd;
			return 0;
		}
	}
	return -1;
}


/*
 *	Activate SLIP on the interface. Lots of messy ioctl calls but the
 *	basic logic is fairly simple.
 *
 *	Returns the name of the SLIP interface allocated on success,
 *	NULL otherwise (and prints a helpful message).
 *	Note that the result returned is a pointer to a static buffer
 *	whose value will be overwritten by subsequent calls.
 */
 
static char *attach_slip(int tty_fd, const char *tty_name/*only for error msgs*/,
	const char *src, const char *dst, int mtu)
{
	static char slip_if_name[32];
	int d;
	int net_fd;
	struct ifreq ifr;
	/* ifr.ifr_addr is actually a struct sockaddr, but that needs to
	 * be cast to the right sort of sockaddr struct if we want to make
	 * sense of its innards.
	 */
	struct sockaddr_in *ifr_addr_in = (struct sockaddr_in *)&ifr.ifr_addr;
	
	/*
	 *	Make this a network interface
	 */
	d=N_SLIP;
	if(ioctl(tty_fd, TIOCSETD, &d)==-1)
	{
		fprintf(stderr,"%s: unable to attach SLIP interface to tty %s\n",program_name,tty_name);
		perror("ioctl(TIOCSETD)");
		return (char*)0;
	}
	
	/*
	 *	Get the network name of the device (eg sl0)
	 */
	if(ioctl(tty_fd, SIOCGIFNAME, slip_if_name)==-1)
	{
		fprintf(stderr,"%s: unable to read SLIP device name\n", program_name);
		perror("ioctl(SIOCGIFNAME)");
		return (char*)0;
	}
	
	/*
	 *	Select basic slip (just for now)
	 */
	d=SL_MODE_SLIP;
	if(ioctl(tty_fd, SIOCSIFENCAP, &d)==-1)
	{
		fprintf(stderr,"%s: %s: unable to select uncompressed SLIP\n", program_name,slip_if_name);
		perror("ioctl(SIOCSIFENCAP)");
		return (char*)0;
	}
	
	/*
	 *	We now need an AF_INET descriptor
	 *	to talk to the networking layer
	 */
	net_fd=socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if(net_fd==-1)
	{
		fprintf(stderr,"%s: unable to create INET socket.\n", program_name);
		perror("socket");
		return (char*)0;
	}
	
	/*
	 *	Set our address
	 */
	strcpy(ifr.ifr_name, slip_if_name);
	ifr.ifr_addr.sa_family=AF_INET;
	if( get_ip_addr(src, &(ifr_addr_in->sin_addr.s_addr)) )
		return (char*)0;
	if(ioctl(net_fd, SIOCSIFADDR, &ifr)==-1)
	{
		fprintf(stderr,"%s: %s: unable to set source IP address.\n",
			program_name,slip_if_name);
		perror("ioctl(SIOCSIFADDR)");
		return (char*)0;
	}

	/*
	 *	Set the target address
	 */
	strcpy(ifr.ifr_name,slip_if_name);
	ifr.ifr_addr.sa_family=AF_INET;
	if( get_ip_addr(dst, &(ifr_addr_in->sin_addr.s_addr)) )
		return (char*)0;
	if(ioctl(net_fd, SIOCSIFDSTADDR, &ifr)==-1)
	{
		fprintf(stderr,"%s: %s: unable to set target IP address.\n",
			program_name,slip_if_name);
		return (char*)0;
	}

	/*
	 *	Set MTU for reasonable network sharing and error tolerance 
	 */
	strcpy(ifr.ifr_name,slip_if_name);
	ifr.ifr_mtu=mtu;
	if(ioctl(net_fd, SIOCSIFMTU, &ifr)==-1)
	{
		fprintf(stderr,"%s: %s: unable to set MTU=%d\n", 
			program_name,slip_if_name,mtu);
		perror("ioctl(SIOCSIFMTU)");
		return (char*)0;
	}
	
	/*
	 *	Retrieve the flags...
	 */
	strcpy(ifr.ifr_name,slip_if_name);
	if(ioctl(net_fd, SIOCGIFFLAGS, &ifr)==-1)
	{
		fprintf(stderr,"%s: %s: unable to read flags.\n",
			program_name, slip_if_name);
		perror("ioctl(SIOCGIFFLAGS)");
		return (char*)0;
	}
	/*
	 *	mark the interface pointopoint and running,
	 *	and set flags in the kernel.
	 */
	strcpy(ifr.ifr_name,slip_if_name);
	ifr.ifr_flags |= IFF_UP | IFF_POINTOPOINT | IFF_RUNNING;
	if(ioctl(net_fd, SIOCSIFFLAGS, &ifr)==-1)
	{
		fprintf(stderr,"%s: %s: unable to set flags.\n",
			program_name, slip_if_name);
		perror("ioctl(SIOCGIFFLAGS)");
		return (char*)0;
	}
	
	/*
	 *	Mission complete
	 */
	return slip_if_name;
}

/*	Create a pty pair and attach a new slip interface to
 *	one side.  If all goes well, return the fd of the other side of
 *	the pty pair; this fd can be used to read and write slip-delimited
 *	frames from/to the kernel.
 */
int kernel_start_slip(const char *local_ip, const char *remote_ip, int mtu,
	char **slip_if_name)
{
	int pty_fd;
	int tty_fd;
	if(ptypair(&pty_fd,&tty_fd)==-1)
	{
		fprintf(stderr,"%s: no pty available\n",program_name);
		return -1;
	}

	if( !(*slip_if_name=attach_slip(
		tty_fd, program_name/*only used for error meaages*/,
		local_ip, remote_ip, mtu)) )
	{
		/* attach_slip will have printed a useful message */
		return -1;
	}

	return (kernel_fd=pty_fd);
}


/* Can change where next frame received is put, if necessary */
void kernel_recv_slip_frame(byte *buf[], int *len)
{
	if(debug & DEBUG_KERNEL)
	{
		fprintf(stderr,"kernel_recv_slip_frame: got %d\n",*len);
		debug_packet(*buf,*len);
	}
	scarab_send_slip_frame(*buf,*len);
}


void kernel_send_slip_frame(byte buf[], int len)
{
	if(debug & DEBUG_KERNEL)
	{
		fprintf(stderr,"kernel_send_slip_frame: sending %d\n",len);
		debug_packet(buf,len);
	}
	slip_write_frame(kernel_fd);/* don't really need this since no noise */
	debug_write(kernel_fd, buf, len);
	slip_write_frame(kernel_fd);
}
