/* Scarab - provides SLIP over a half-duplex link such as infrared or radio.
 *
 * main.c
 * Converts command line options in to settings held in global variables; 
 * main loop receieving and sending packets in both directions.
 *
 * (c) Copyright 1996 Alan Cox, Robin O'Leary <robin@acm.org>.
 * All Rights Reserved.
 *
 * This program is distributed without any warranty; without even the
 * implied warranty of MERCHANTABILITY or fitness for a particular purpose.
 * See the file LICENSE for details.
 */

#include <stdio.h>	/* fprintf() */
#include <getopt.h>	/* getopt() */
#include <unistd.h>	/* fork() */
#include <sys/time.h>	/* struct timeval */

#include "main.h"
#include "debug.h"
#include "kernel.h"
#include "scarab.h"
#include "slip.h"
#include "version.h"
#include "date.h"

#define MAX(A,B) ( (A) > (B) ? (A) : (B) )


/* All times are in microseconds, but milliseconds or seconds are more
 * convenient for humans, so provide conversions.
 */
#define US 1
#define MS 1000
#define S 1000000

/* 2nd argument of show_options() is a set of bit flags */
#define SHOW_USAGE	1
#define SHOW_SETTINGS	2

/******************************************************************************
 *
 * Configurable things and global settings
 */

/* RFC 1066 suggests a limit of around 1006, so this should be plenty */
#define SLIP_BUFFER_SIZE	4020

/* In order to share the channel fairly, we have some timeouts which give
 * the other side a look-in.
 */
/* SCARAB_WAIT_FOR_MORE is the time we wait after data has come in to see
 * if there is any more.  This prevents us jumping on the tail end of a
 * packet, or in a multi-node system, allows other nodes to get in before
 * us.  In such a system, this delay should be dynamic.
 */
#ifdef SLOW
int scarab_wait_for_more=200*MS;
#else
int scarab_wait_for_more=1*MS;
#endif

/* KERNEL_WAIT_AFTER_TX forces a dead time after each of our transmission
 * to give the other side a chance to start up if it wants to.
 */
#ifdef SLOW
int kernel_wait_after_tx=400*MS;
#else
int kernel_wait_after_tx=2*MS;
#endif

int scarab_has_dcd=0;
int scarab_has_rts=0;
int scarab_is_airport=0;
int scarab_baud=9600;
int scarab_mtu=300;
int scarab_train_len=10;
int scarab_train_byte=0xff;
const char *scarab_port="/dev/scarab";
int scarab_needs_node_id=0;
byte scarab_node_id=0;
const char *local_ip="no default";
const char *remote_ip="no default";

/* If scarab_has_rts is set, then we raise RTS while we are sending
 * data.  We allow some extra time before and after the data to
 * give the tranmitter a chance to power up, and to give the last
 * byte time to get sent.
 */
#ifdef SLOW
int scarab_rts_start=   5*MS;	/* time tx takes to power up */
int scarab_rts_stop = 500*US;	/* 1 char =~ 100us at 9600 baud */
#else
int scarab_rts_start= 5*US;	/* time tx takes to power up */
int scarab_rts_stop = 1*US;	/* 1 char =~ 1us at 1M baud */
#endif

int scarab_is_transmitting=0; /* flag set by scarab.c */

int debug=0;

const char *program_name;


/******************************************************************************
 *
 * Control code
 */

void print_options(FILE *stream, int detail)
{
	if(detail & SHOW_USAGE)
	{
		fprintf(stream,"Usage: %s [-hVadr] [-b baud] [-m mtu] [-n nodeID] [-p port] [-l trainlen] [-t trainbyte] [-D debug] local-addr remote-addr\n",program_name);
	}
	if(detail & SHOW_SETTINGS)
	{
		fprintf(stream,"    -a          IrDA link needs AirPort commands [%s]\n", scarab_is_airport?"yes":"no");
		fprintf(stream,"    -b baud     speak to modem at this rate [%d]\n", scarab_baud);
		fprintf(stream,"    -d          modem is equipped with DCD [%s]\n", scarab_has_dcd?"yes":"no");
		fprintf(stream,"    -D debug    enable debug messages [0x%03x] (-D? for help)\n", debug);
		fprintf(stream,"    -h          display this help message (also shows option settings)\n");
		fprintf(stream,"    -l len      length of training data [%d]\n",scarab_train_len);
		fprintf(stream,"    -m mtu      maximum packet size [%d]\n", scarab_mtu);
		fprintf(stream,"    -n nodeID   node ID for Scarab models that need it or -1 [%d%s]\n", scarab_needs_node_id?scarab_node_id:-1, scarab_needs_node_id?" (not needed)":"");
		fprintf(stream,"    -p port     tty connected to scarab modem [%s]\n", scarab_port);
		fprintf(stream,"    -r          modem needs RTS to tell it to transmit [%s]\n", scarab_has_rts?"yes":"no");
		fprintf(stream,"    -t byte/-pattern\n");
		fprintf(stream,"                training data [%s %d]\n",(scarab_train_byte<0)?"pattern":"byte",scarab_train_byte);
		fprintf(stream,"    -V          display version information\n");
		fprintf(stream,"    local-addr  IP address or hostname of local machine [%s]\n",local_ip);
		fprintf(stream,"    remote-addr IP address or hostname of remote SLIP server [%s]\n",remote_ip);
	}
	else
	{
		fprintf(stream,"For more help type: %s -h\n",program_name);
	}
}

void print_version(FILE *f, int detail)
{
	if(detail)
	{
		fprintf(f,"%s: Scarab node controller for Linux\n",program_name);
		fprintf(f,"Version %s, compiled %s.\n",version,date);
	}
	else
	{
		fprintf(f,"scarabd version %s\n",version);
	}
	fprintf(f,"Copyright (C) 1996 Alan Cox, Robin O'Leary <robin@acm.org>. All rights reserved.\n");
}


void server(int scarab_fd, int kernel_fd)
{
	struct slip_state *scarab_slip;
	struct slip_state *kernel_slip;
	byte scarab_buffer[SLIP_BUFFER_SIZE];
	byte kernel_buffer[SLIP_BUFFER_SIZE];
	int max_fd=MAX(scarab_fd,kernel_fd)+1;
	enum { IDLE, TX, RX, WAIT } state=IDLE;
	static const char *state_name[]={"idle","tx","rx","wait"};

	scarab_slip=slip_make_state(
		scarab_buffer, SLIP_BUFFER_SIZE,
		scarab_fd, scarab_recv_slip_frame);
	kernel_slip=slip_make_state(
		kernel_buffer, SLIP_BUFFER_SIZE,
		kernel_fd, kernel_recv_slip_frame);
	for(;;)
	{
		fd_set read_fds;
		struct timeval timeout;
		int s;

		FD_ZERO(&read_fds);
		timeout.tv_sec=0;
		timeout.tv_usec=0;

		if(debug & DEBUG_LIGHTS)
		{
			static byte count=0;
			fprintf(stderr,"|%02x|%s|%s|\r",
				count++,
				(state==TX)?"TX":(state==WAIT)?"T ":"  ",
				(state==RX)?"RX":"  "
			);
		}
		switch(state)
		{
			case IDLE: /* nothing happening */
				FD_SET(kernel_fd, &read_fds);
				FD_SET(scarab_fd, &read_fds);
				timeout.tv_sec=999;
				break;
			case TX: /* currently transmitting */
				FD_SET(kernel_fd, &read_fds);
				break;
			case RX: /* currently receiving */
				FD_SET(scarab_fd, &read_fds);
				timeout.tv_usec=scarab_wait_for_more;
				break;
			case WAIT: /* just finished transmitting */
				FD_SET(scarab_fd, &read_fds);
				timeout.tv_usec=kernel_wait_after_tx;
				break;
		}

		if(debug & DEBUG_STATE)
		{
			fprintf(stderr,"server: %s - waiting %ld%s for%s%s...\n",
				state_name[state],
				timeout.tv_sec?timeout.tv_sec:(timeout.tv_usec/1000),
				timeout.tv_sec?"s":"ms",
				FD_ISSET(scarab_fd, &read_fds)?" scarab":"",
				FD_ISSET(kernel_fd, &read_fds)?" kernel":""
			);
		}

		if(debug & DEBUG_STATE)
		    fprintf(stderr,"server: select(%d,0x%lx,NULL,NULL,%d.%06d)",
			    max_fd, read_fds.fds_bits[0], timeout.tv_sec, timeout.tv_usec);
		s=select(max_fd, &read_fds, NULL, NULL, &timeout);
		if(debug & DEBUG_STATE)
		    fprintf(stderr,"=%d (0x%lx)\n",s,read_fds.fds_bits[0]);
		switch(s)
		{
			case -1:
			{
				fprintf(stderr,"server: bad select\n");
				perror("select");
				break;
			}
			case 0:
			{
				if(debug & DEBUG_STATE)
				{
					fprintf(stderr,"server: - timeout\n");
				}
				switch(state)
				{
					case IDLE:
						break;
					case TX:
						scarab_stop_transmitting();
						if(debug & DEBUG_STATE)
							fprintf(stderr,"server: - stopped transmitting; waiting\n");
						state=WAIT;
						break;
					case RX:
					case WAIT:
						if(debug & DEBUG_STATE)
							fprintf(stderr,"server: - idle\n");
						state=IDLE;
						break;
				}
				break;
			}
		    case 1:
		    case 2:
			if(debug & DEBUG_STATE)
			{
				fprintf(stderr,"server: - got data from%s%s\n",
					FD_ISSET(scarab_fd, &read_fds)?" scarab":"",
					FD_ISSET(kernel_fd, &read_fds)?" kernel":""
				);
			}
			if(FD_ISSET(scarab_fd, &read_fds))
			{
				slip_stream(scarab_slip);
				if(debug & DEBUG_STATE)
					fprintf(stderr,"server: - receiving\n");
				state=RX;
			}
			else
			if(FD_ISSET(kernel_fd, &read_fds))
			{
				slip_stream(kernel_slip);
				if(scarab_is_transmitting)
				{
					if(debug & DEBUG_STATE)
						fprintf(stderr,"server: - transmitting\n");
					state=TX;
				}
			}
			break;
		}/*switch(select)*/
	}/*forever*/
}

int main(int argc, char *argv[])
{
	char *slip_if_name;
	int c;
	int errflg=0;
	int kernel_fd=0;
	int scarab_fd=0;
	int pid;

	program_name=argv[0];

	while ((c = getopt(argc, argv, "ab:dD:hl:m:n:p:rt:V")) != EOF) {
		switch (c) {
			case 'a': {
				scarab_is_airport=1;
				break;
			}
			case 'b': {
				if(sscanf(optarg,"%d",&scarab_baud)!=1)
				{
					fprintf(stderr,"%s: invalid baud `%s'\n", program_name,optarg);
					errflg++;
				}
				break;
			}
			case 'd': {
				scarab_has_dcd=1;
				break;
			}
			case 'D': {
				if(!strcmp(optarg,"?") || !strcmp(optarg,"h"))
				{
					printf("%s: -D debug levels are:\n"
						"DEBUG_TERM	0x%04x  Trace tty set-up\n"
						"DEBUG_READ	0x%04x  Report bytes received\n"
						"DEBUG_WRITE	0x%04x  Report bytes sent\n"
						"DEBUG_BYTE	0x%04x  Enable byte-level display\n"
						"DEBUG_SLIP 	0x%04x  Report status of slip frames\n"
						"DEBUG_KERNEL	0x%04x  Show traffic to/from kernel\n"
						"DEBUG_SCARAB	0x%04x  Show traffic to/from Scarab modem\n"
						"DEBUG_STATE	0x%04x  High-level status report\n"
						"DEBUG_LIGHTS	0x%04x  Very abbreviated status display\n"
						,program_name
						,DEBUG_TERM
						,DEBUG_READ
						,DEBUG_WRITE
						,DEBUG_BYTE
						,DEBUG_SLIP
						,DEBUG_KERNEL
						,DEBUG_SCARAB
						,DEBUG_STATE
						,DEBUG_LIGHTS
					);
					exit(0);
				}
				else
				if(sscanf(optarg,"%i",&debug)!=1)
				{
					fprintf(stderr,"%s: invalid debug level `%s'\n", program_name,optarg);
					errflg++;
				}
				break;
			}
			case 'h': {
				print_options(stdout,SHOW_USAGE | SHOW_SETTINGS);
				exit(0);
			}
			case 'l': {
				if(sscanf(optarg,"%d",&scarab_train_len)!=1)
				{
					fprintf(stderr,"%s: invalid training length `%s'\n", program_name,optarg);
					errflg++;
				}
				if(scarab_train_len==-1)
				{
					scarab_train_len=0;
					scarab_train_byte=-1;
				}
				if(scarab_train_len<0 || scarab_train_len>1024)
				{
					fprintf(stderr,"%s: training length must be between 0 and 1024\n",
							program_name);
					errflg++;
				}
				break;
			}
			case 'm': {
				if(sscanf(optarg,"%d",&scarab_mtu)!=1)
				{
					fprintf(stderr,"%s: invalid mtu `%s'\n", program_name,optarg);
					errflg++;
				}
				break;
			}
			case 'n': {
				if(sscanf(optarg,"%i",&c)!=1)
				{
					fprintf(stderr,"%s: invalid node ID `%s'\n", program_name,optarg);
					errflg++;
				}
				else
				if(c<-1 || c>255)
				{
					fprintf(stderr,"%s: node ID must be between -1 and 255\n", program_name);
					errflg++;
				}
				else
				if(c==-1)
				{
					scarab_needs_node_id=scarab_node_id=0;
				}
				else
				{
					scarab_needs_node_id=1;
					scarab_node_id=c;
				}
				break;
			}
			case 'p': {
				scarab_port=optarg;
				break;
			}
			case 'r': {
				scarab_has_rts=1;
				break;
			}
			case 't': {
				if(sscanf(optarg,"%i",&scarab_train_byte)!=1)
				{
					fprintf(stderr,"%s: invalid training byte `%s'\n", program_name,optarg);
					errflg++;
				}
				else
				if(scarab_train_byte<-1 || scarab_train_byte>255)
				{
					fprintf(stderr,"%s: training byte must be between -1 and 255\n",
							program_name);
					errflg++;
				}
				break;
			}
			case 'V': {
				print_version(stdout,1/*long*/);
				exit(0);
			}
			case '?': {
				errflg++;
				break;
			}
		}
	}

	if( (argc-optind) == 2)
	{
		local_ip=argv[optind+0];
		remote_ip=argv[optind+1];
	}
	else
	{
		fprintf(stderr,"%s: not enough arguments\n", program_name);
		errflg++;
	}

	if(errflg)
	{
		print_options(stderr,SHOW_USAGE);
		exit(1);
	}

	print_version(stderr,debug);

	scarab_fd=scarab_open(scarab_port, scarab_baud);
	if(scarab_fd==-1)
		exit(1);
	
	kernel_fd=kernel_start_slip(local_ip, remote_ip, scarab_mtu, &slip_if_name);
	if(kernel_fd==-1)
		exit(1);

	if(debug)
	{
		fprintf(stderr,"Debug %#04x enabled; current status:\n",debug);
		print_options(stdout,SHOW_SETTINGS);
		fprintf(stderr,"scarab fd %d, kernel fd %d\n",
			scarab_fd,kernel_fd);
		fprintf(stderr,"Starting child process...\n");
	}

	/* Put server process in background */
	switch(pid=fork())
	{
		case -1:
			perror("fork");
			exit(1);
		case 0:
			server(scarab_fd, kernel_fd);
			break;
		default:
			printf("%s %d\n",slip_if_name,pid);
			break;
	}
	return 0;
}
