/* Scarab - provides SLIP over a half-duplex link such as infrared or radio.
 *
 * scarab.c
 * Handles packets going to and coming from the half-duplex device.
 * Does training, timimg and switching of device, and framing of packet.
 *
 * (c) Copyright 1996 Alan Cox, Robin O'Leary <robin@acm.org>.
 * All Rights Reserved.
 *
 * This program is distributed without any warranty; without even the
 * implied warranty of MERCHANTABILITY or fitness for a particular purpose.
 * See the file LICENSE for details.
 */

#include <sys/ioctl.h>	/* ioctl() */
#include <unistd.h>	/* usleep() */
#include <stdio.h>	/* perror() */
#include <string.h>	/* memset() */
#include <fcntl.h>	/* open() */
#include <sys/time.h>	/* struct timeval */
#include <termios.h>	/* */

#include "main.h"	/* types, globals */
#include "debug.h"	/* for DEBUG_SCARAB */
#include "term.h"	/* for set_termios() */
#include "slip.h"	/* for slip_write_frame() */
#include "kernel.h"	/* for kernel_send_slip_frame() */
#include "airport.h"	/* for airport_init() */
#include "scarab.h"	/* consistency check */

#define BUFFERED_WRITE_LENGTH	4000


/* If we want to send something, but discover that the carrier detect
 * says that the channel is in use, we wait a while before looking again.
 * This delay should get bigger the busier the channel is, but at least
 * in version 1.2, it is fixed.  It hasn't mattered yet, because we have
 * only ever had two radios to play with, so contention hasn't been an
 * issue!  Time is in microseconds.
 */
#define BACKOFF_WAIT		100000


/* This is probably a bad idea.  How will we cope when we have more
 * than one modem?
 */
static int scarab_fd=-1;



int buffered_write(int fd, const char *data, int len)
{
	static char buffer[BUFFERED_WRITE_LENGTH];
	static int buffer_count=0;

	if(len)
	{
		if(len+buffer_count>=BUFFERED_WRITE_LENGTH)
		{
			fprintf(stderr,"buffered_write: too much data (%d new, %d already)\n",len,buffer_count);
			return -1;
		}
		memcpy(buffer+buffer_count,data,len);
		buffer_count+=len;
		return 0;
	}
	else
	{
		int result;
		if(debug & DEBUG_SCARAB)
		{
			fprintf(stderr,"buffered_write: sending %d bytes\n",buffer_count);
			debug_packet(buffer,buffer_count);
		}
		result=debug_write(fd,buffer,buffer_count);
		buffer_count=0;
		return result;
	}
}

/********************** Private functions ***********************/
 
static void scarab_assert_rts(void)
{
	if(debug & DEBUG_SCARAB)
		fprintf(stderr,"scarab_assert_rts\n");
	if(set_termios_rts(scarab_fd,1))
	{
		fprintf(stderr,"%s: can't assert RTS; disabling -r\n",
			program_name);
		scarab_has_rts=0;
	}
	usleep(scarab_rts_start);	/* Give tx time to start up */
}

static void scarab_release_rts(void)
{
	if(debug & DEBUG_SCARAB)
		fprintf(stderr,"scarab_release_rts\n");
	tcdrain(scarab_fd);	/* Wait for data to leave Linux */
	/* don't bother; we sent an extra character so we don't
	 * mind if it gets cut off
	usleep(scarab_rts_stop);	/* ...and a while for radio to transmit it */
	if(set_termios_rts(scarab_fd,0))
	{
		fprintf(stderr,"%s: can't release RTS; disabling -r\n",
			program_name);
		scarab_has_rts=0;
	}
}


static void scarab_train(void)
{
	if(debug & DEBUG_SCARAB)
		fprintf(stderr,"scarab_train\n");
	switch(scarab_train_byte)
	{
		case -1:
		{
			char buf[256];
			memset(buf,0x55,128);
			memset(buf+128,0xAA,128);
			buffered_write(scarab_fd,buf,256);
			break;
		}
		default:
		{
			byte b=scarab_train_byte;
			int i;
			for(i=0; i<scarab_train_len; i++)
				buffered_write(scarab_fd,&b,1);
			break;
		}
	}
}

static void scarab_wait_clear(void)
{
	int v;
	int s=0;
	if(debug & DEBUG_SCARAB)
		fprintf(stderr,"scarab_wait_clear...\n");
	do {
		/*
		 *	This is crude, we just backoff .1 of a second if
		 *	the link is busy. We can use s to do p-persist
		 *	algorithms if we have >2 units.
		 */
		if(s++>0)
			usleep(BACKOFF_WAIT);
		ioctl(scarab_fd, TIOCMGET,&v);
	} while(!(v&TIOCM_CAR));	/* Wait for DCD to drop */

	if(debug & DEBUG_SCARAB)
		fprintf(stderr,"...clear\n");
}

/********************** Scarab radio modem library ***********************/


int scarab_open(const char *filename, int baud)
{
	if(debug & DEBUG_SCARAB)
		fprintf(stderr,"scarab_open: opening %s at %d baud\n",
			filename,baud);
	scarab_fd=open(filename, O_RDWR);
	if(scarab_fd==-1)
	{
		perror(filename);
		return -1;
	}
	
	if(scarab_is_airport)
	{
		if(debug & DEBUG_SCARAB)
			fprintf(stderr,"scarab_open: initialising AirPort\n");
		if(airport_init(scarab_fd,baud))
		{
			fprintf(stderr,"scarab_open: AirPort init failed.\n");
			return -1;
		}
	}

	if(set_termios(scarab_fd,baud))
		return -1;	/* set_termios prints a mesasge */

	if(debug & DEBUG_SCARAB)
		fprintf(stderr,"scarab_open: open ``%s'' on fd %d\n",filename,scarab_fd);

	return scarab_fd;
}


/* Can change where next frame received is put, if necessary */
void scarab_recv_slip_frame(byte *buf[], int *len)
{
	if(debug & DEBUG_SCARAB)
	{
		fprintf(stderr,"scarab_recv_slip_frame: got %d\n",*len);
		debug_packet(*buf,*len);
	}
        if(*len<20)
        {
                if(debug & DEBUG_SCARAB)
                        fprintf(stderr,"-ignored because it's too short\n");
		return;
        }

	/* ignore our own packets */
	if(scarab_needs_node_id)
	{
		if(*buf[0]==scarab_node_id)
		{
			if(debug & DEBUG_SCARAB)
				fprintf(stderr,"-ignored because it's our id %d\n",scarab_node_id);
		}
		else
		{
			kernel_send_slip_frame(*buf+1, *len-1);
		}
	}
	else
	{
		kernel_send_slip_frame(*buf, *len);
	}
}


void scarab_send_slip_frame(byte buf[], int len)
{
	if(!scarab_is_transmitting)
	{
		if(debug & DEBUG_SCARAB)
		{
			fprintf(stderr,"scarab_send_slip_frame: starting tx\n");
		}
		/* Either transmit when carrier says it's clear, or just wait
		 * a while and hope.
		 */
		if(scarab_has_dcd)
			scarab_wait_clear();

		if(scarab_has_rts)
			scarab_assert_rts();	/* Grab it */
		scarab_train();		/* Train modem, send header */

		/* Remember that the modem is in use */
		/* this is all we need, but...
		scarab_is_transmitting=1;
		*/
	}
	/* ...it might be nice to count packets */
	scarab_is_transmitting++;

	if(debug & DEBUG_SCARAB)
	{
		fprintf(stderr,"scarab_send_slip_frame: sending %d\n",len);
		debug_packet(buf,len);
	}
	buffered_write(scarab_fd, slip_frame_marker(),1);
	if(scarab_needs_node_id)
		buffered_write(scarab_fd,&scarab_node_id,1); /* Send node id */
	buffered_write(scarab_fd,buf,len);
	buffered_write(scarab_fd, slip_frame_marker(),1);
#ifdef TRANSMIT_TWICE
	if(scarab_needs_node_id)
		buffered_write(scarab_fd,&scarab_node_id,1); /* Send node id */
	buffered_write(scarab_fd,buf,len);
	buffered_write(scarab_fd, slip_frame_marker(),1);
#endif
}

void scarab_stop_transmitting(void)
{
	buffered_write(scarab_fd, slip_frame_marker(),1); /* dummy byte */
	buffered_write(scarab_fd,0,0);	/* Send data */
	if(debug & DEBUG_SCARAB)
	{
		fprintf(stderr,"scarab_stop_transmitting: sent %d frames\n",
			scarab_is_transmitting);
	}
	if(scarab_has_rts)
		scarab_release_rts();	/* Release the link */
	scarab_is_transmitting=0;
}
