/* Scarab - provides SLIP over a half-duplex link such as infrared or radio.
 *
 * slip.c
 * Basic SLIP state machine.
 *
 * (c) Copyright 1996 Alan Cox, Robin O'Leary <robin@acm.org>.
 * All Rights Reserved.
 *
 * This program is distributed without any warranty; without even the
 * implied warranty of MERCHANTABILITY or fitness for a particular purpose.
 * See the file LICENSE for details.
 */

#include <stdio.h>	/* fprintf() */
#include <stdlib.h>	/* calloc() */
#include "main.h"	/* types, globals */
#include "debug.h"	/* for DEBUG_SLIP */
#include "slip.h"	/* consistency check */


/* Magic numbers in SLIP frames */
#define SLIP_END	0300
#define SLIP_ESC	0333
#define SLIP_ESC_ESC	0334
#define SLIP_ESC_END	0335
 

/* could do byte-by-byte read since we look at each byte anyway, but
 * maybe this is a bit more efficient
 */
#define READ_BUFFER_SIZE	2000

enum ss {
	SS_HUNT,		/* looking for start of frame */
	SS_DATA,		/* reading data from frame */
	SS_ESCAPE		/* next byte read has been escaped */
};


struct slip_state
{
	enum ss rx_state;	/* track state of SLIP framing */
	int rx_buflen;		/* length of byte receive buffer */
	byte *rx_buffer;	/* receive buffer; allocated by caller */
	int rx_bcount;		/* bytes in receive buffer */
	int rx_fd;		/* where to get bytes */
	void (*rx_frame)(byte *buf[], int *len); /* call when frame received */
};


struct slip_state *slip_make_state(
	byte *buffer, int buflen,
	int rx_fd,
	void (*rx_frame)(byte *buf[], int *len) )
{
	struct slip_state *new=(struct slip_state *)calloc(1,sizeof(struct slip_state));
	if(new)
	{
		new->rx_state=SS_HUNT;
		new->rx_bcount=0;
		new->rx_buflen=buflen;
		new->rx_buffer=buffer;
		new->rx_fd=rx_fd;
		new->rx_frame=rx_frame;
	}
	return new;
}

/*
 *	Note: This doesn't implement SLIP framing, it understands it.  The
 *	resulting buffer is an intact SLIP frame with frame markers deleted.
 *
 *	Returns:
 *	   0	still waiting for a frame
 *	  -1	invalid byte or frame too big; frame discarded
 *	   1	valid frame received.
 */
static int slip_state_machine(struct slip_state *sl, unsigned char byte)
{
	switch(sl->rx_state)
	{
	case SS_HUNT:	/* looking for frame start */
	{
		if(byte==SLIP_END) /* Got it.  Start recording data */
		{
			sl->rx_state=SS_DATA;
			sl->rx_bcount=0;
		}
		return 0;
	}
	case SS_ESCAPE:
	{
		/* Only two values need to be escaped;
		 * if we see anything else it's an error.
		 */
		if(byte!=SLIP_ESC_ESC && byte!=SLIP_ESC_END)
		{
			sl->rx_state=SS_HUNT;
			return -1;	
		}

		/* Note!  We don't un-escape the bytes; we just
		 * store them.  That way, when we're done, there is
		 * a valid SLIP packet sitting in the buffer ready
		 * to be retransmitted.
		 */
		sl->rx_state=SS_DATA;
		break;
	}
	case SS_DATA:
	{
		if(byte==SLIP_ESC)
			sl->rx_state=SS_ESCAPE;
		else
		if(byte==SLIP_END)
		{
			if(sl->rx_bcount)	/* Discard empties */
			{
				if(debug & DEBUG_SLIP)
				{
					fprintf(stderr,"slip_state_machine: got %d byte packet from fd %d\n",sl->rx_bcount,sl->rx_fd);
					debug_packet(sl->rx_buffer,sl->rx_bcount);
				}
				sl->rx_frame(&sl->rx_buffer, &sl->rx_bcount);
			}
			
			/* Stay in SS_DATA mode ready for next frame */
			sl->rx_bcount=0;
			return 0;
		}
		break;
	}
	}
	
	/* If we get here, we are supposed to add the byte to the
	 * slip received buffer.
	 */
	
	/*
	 *	Never let buffer overflow
	 *	Just throw away oversize packet
	 */
	if(sl->rx_bcount==sl->rx_buflen)
	{
		sl->rx_state=SS_HUNT;
		return -1;	
	}
	sl->rx_buffer[sl->rx_bcount++]=byte;
	
	return 0;
}

void slip_stream(struct slip_state *sl)
{
	byte c[READ_BUFFER_SIZE];
	int l;
	if((l=debug_read(sl->rx_fd, c,READ_BUFFER_SIZE))>0)
	{
		char *p=&c[0];
		if(debug & DEBUG_SLIP)
		{
			fprintf(stderr,"slip_stream(%d): read %d bytes\n",
				sl->rx_fd, l);
		}
		while(l--)
			slip_state_machine(sl,*p++);
	}
}

const byte *slip_frame_marker(void)
{
	static const byte b=SLIP_END;
	return &b;
}

void slip_write_frame(int fd)
{
	debug_write(fd,slip_frame_marker(),1);
}
